'******************************************************************************
' Copyright (c) 2011, 2012, Dave Hein
' See end of file for terms of use.
'******************************************************************************
CON
  SEEK_SET       = 0
  SEEK_END       = 2
  KILL_COMMAND   = 25
  STACK_SIZE     = 800
  START_ADDR     = $20
  FILE_NOT_EXIST = -3
  OUT_OF_MEMORY  = -2

  ' LMM Kernel Registers
  r0 = 0
  r1 = 1
  r2 = 2
  r3 = 3
  r4 = 4
  r5 = 5
  lr = 15
  pc = 17

OBJ
  c   : "cfileiosd"
  mem : "cmalloc"
  sys : "sysdefs"

' This routine loads and runs a C program from a file  
PUB run(argc, argv, mode) | cognum, infile, size, i, size1, argsize, argv1, ptr

  ' Open the program file
  infile := c.fopen(long[argv], string("r"))
  ifnot infile
    return FILE_NOT_EXIST

  ' Determine the size of the arg list
  argsize := 4 ' Initialize argsize to 4 bytes for NULL argv terminator
  repeat i from 0 to argc - 1
    argsize += 4 + strsize(long[argv][i]) + 1
  argsize := ((argsize + 3) >> 2) << 2 ' Round to the nearest long

  if mode & sys#run_stand_alone
    return c.run_stand_alone(argc, argv, argsize, infile, mode)

  ' Determine the file size
  c.fseek(infile, 0, SEEK_END)
  size := c.ftell(infile)
  c.fseek(infile, 0, SEEK_SET)
  size := ((size + 3) >> 2) << 2 ' Round to the nearest long

  ' Kill the calling program if requested to free up memory
  if (mode & sys#run_kill_caller)
    argv := c.KillCaller(argc, argv, argsize, mode)

  ' Allocate memory for the app plus space for the stack and the arg list
  size1 := mem.malloc0(size + STACK_SIZE + argsize)
  ifnot size1
    c.fclose(infile)
    return OUT_OF_MEMORY

  ' Skip the first 6 bytes and read the app file into memory
  c.fread($6, 1, 6, infile)
  c.fread($6, 1, size - 6, infile)
  c.fclose(infile)

  ' Patch cog location $2 to test if lock has been allocated
  ' instead of testing for stack address PAR not equal to $8000
  long[START_ADDR][r2] := patch | (long[START_ADDR][r5] & $1ff)

  ' Copy arg list to app after the stack
  argv1 := size + STACK_SIZE
  c.copyarglist(argc, argv, argv1)
  if (mode & sys#run_kill_caller)
    mem.free(argv)

  ' Patch the address of the arg list into the app
  ' The pc contains the entry address, and the third long after
  ' the entry address points to the arg list
  ptr := long[START_ADDR][pc]
  long[ptr][3] := argv1

  ' Patch in the address of the exit code
  ' The lr register contains the address of the exit routine, and
  ' the fist word after the start of the exit routine is the address
  ' of the exit hook routine.
  if (mode & sys#run_stand_alone)
    argv := @stand_alone_exit
  else
    argv := @spinix_app_exit
  ptr := long[START_ADDR][lr]
  long[ptr][1] := argv

  ' If running stand-alone stop all cogs, clear and return locks and
  ' start the app in cog 0
  if (mode & sys#run_stand_alone)
    repeat i from 0 to 7
      if i <> cogid
        cogstop(i)
    repeat i from 0 to 7
      lockclr(i)
      lockret(i)
    coginit(0, START_ADDR, size + STACK_SIZE)
    cogstop(cogid)

  ' Otherwise, start the app in the next available cog and add an entry
  ' to the cog table
  else
    cognum := cognew(START_ADDR, size + STACK_SIZE)
    if (mode & sys#run_kill_caller) or not (mode & sys#run_shell_nowait)
      long[sys#focus] := cognum
    i := sys#proc_type_capp | ((mode & 7) << 4)
    c.AddToCogTable(cognum, i, long[argv1], 0, size1)

  return cognum

DAT
' This is used to replace an instruction in the LMM kernel so that we can
' use a stack address other than $8000.  The source address for __C_LOCK_PTR
' will be copied from the instruction at cog location 5.
patch   rdlong r1, 0 wz

' This exit routine is used to terminate stand-alone programs by rebooting
stand_alone_exit 
        mov    r1, #128
        clkset r1

' This exit routine is used to terminate spinix apps by issuing a kill command
spinix_app_exit
        ' Get the file lock number and set the lock
        rdlong  r3, pc
        long    sys#filelock
        rdlong  r3, r3
        sub     r3, #1
        lockset r3  wc
  if_c  sub     pc, #8

        ' Get the cog number and save it in hub RAM
        cogid   r0
        mov     r1, pc
        long    0
        wrlong  r0, r1
        
        ' Save the hub RAM address in the file parm location
        rdlong  r5, pc
        long    sys#fileparm
        wrlong  r1, r5

        ' Combine the cog number with the spinix kill command
        shl     r0, #16
        or      r0, #KILL_COMMAND

        ' Write the command to the file command location
        rdlong  r4, pc
        long    sys#filecmd
        wrlong  r0, r4

        ' Spin in a tight loop waiting for the process to terminate
        sub     pc, #4

{{
+-----------------------------------------------------------------------------+
|                       TERMS OF USE: MIT License                             |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
}}
